import { copyFixtureIntoTempDir } from "jest-fixtures";
import {
  ReleasePlan,
  Config,
  NewChangeset,
  ComprehensiveRelease
} from "@changesets/types";
import * as git from "@changesets/git";
import fs from "fs-extra";
import path from "path";
import outdent from "outdent";
import spawn from "spawndamnit";

import applyReleasePlan from "./";
import { getPackages } from "@manypkg/get-packages";

class FakeReleasePlan {
  changesets: NewChangeset[];
  releases: ComprehensiveRelease[];
  config: Config;

  constructor(
    changesets: NewChangeset[] = [],
    releases: ComprehensiveRelease[] = []
  ) {
    const baseChangeset: NewChangeset = {
      id: "quick-lions-devour",
      summary: "Hey, let's have fun with testing!",
      releases: [{ name: "pkg-a", type: "minor" }]
    };
    const baseRelease: ComprehensiveRelease = {
      name: "pkg-a",
      type: "minor",
      oldVersion: "1.0.0",
      newVersion: "1.1.0",
      changesets: ["quick-lions-devour"]
    };
    this.config = {
      changelog: false,
      commit: false,
      linked: [],
      access: "restricted",
      baseBranch: "master"
    };

    this.changesets = [baseChangeset, ...changesets];
    this.releases = [baseRelease, ...releases];
  }

  getReleasePlan(): ReleasePlan {
    return {
      changesets: this.changesets,
      releases: this.releases,
      preState: undefined
    };
  }
}

async function testSetup(
  fixtureName: string,
  releasePlan: ReleasePlan,
  config?: Config,
  setupFunc?: (tempDir: string) => Promise<any>
) {
  if (!config) {
    config = {
      changelog: false,
      commit: false,
      linked: [],
      access: "restricted",
      baseBranch: "master"
    };
  }
  let tempDir = await copyFixtureIntoTempDir(__dirname, fixtureName);
  if (setupFunc) {
    await setupFunc(tempDir);
  }

  if (config.commit) {
    await spawn("git", ["init"], { cwd: tempDir });
    await git.add(".", tempDir);
    await git.commit("first commit", tempDir);
  }

  return {
    changedFiles: await applyReleasePlan(
      releasePlan,
      await getPackages(tempDir),
      config
    ),
    tempDir
  };
}

describe("apply release plan", () => {
  describe("versioning", () => {
    it("should update a version for one package", async () => {
      const releasePlan = new FakeReleasePlan();
      let { changedFiles } = await testSetup(
        "simple-project",
        releasePlan.getReleasePlan(),
        releasePlan.config
      );
      let pkgPath = changedFiles.find(a => a.endsWith("pkg-a/package.json"));

      if (!pkgPath) throw new Error(`could not find an updated package json`);
      let pkgJSON = await fs.readJSON(pkgPath);

      expect(pkgJSON).toMatchObject({
        name: "pkg-a",
        version: "1.1.0"
      });
    });
    it("should not update ranges set to *", async () => {
      const releasePlan = new FakeReleasePlan(
        [
          {
            id: "some-id",
            releases: [{ name: "pkg-b", type: "minor" }],
            summary: "a very useful summary"
          }
        ],
        [
          {
            changesets: ["some-id"],
            name: "pkg-b",
            newVersion: "1.1.0",
            oldVersion: "1.0.0",
            type: "minor"
          }
        ]
      );
      let { changedFiles } = await testSetup(
        "simple-star-dep",
        releasePlan.getReleasePlan(),
        releasePlan.config
      );
      let pkgPath = changedFiles.find(a => a.endsWith("pkg-a/package.json"));

      if (!pkgPath) throw new Error(`could not find an updated package json`);
      let pkgJSON = await fs.readJSON(pkgPath);

      expect(pkgJSON).toEqual({
        name: "pkg-a",
        version: "1.1.0",
        dependencies: {
          "pkg-b": "*"
        }
      });
    });
    it("should update a version for two packages with different new versions", async () => {
      const releasePlan = new FakeReleasePlan(
        [],
        [
          {
            name: "pkg-b",
            type: "major",
            oldVersion: "1.0.0",
            newVersion: "2.0.0",
            changesets: []
          }
        ]
      );

      let { changedFiles } = await testSetup(
        "simple-project",
        releasePlan.getReleasePlan(),
        releasePlan.config
      );
      let pkgPathA = changedFiles.find(a => a.endsWith("pkg-a/package.json"));
      let pkgPathB = changedFiles.find(b => b.endsWith("pkg-b/package.json"));

      if (!pkgPathA || !pkgPathB) {
        throw new Error(`could not find an updated package json`);
      }
      let pkgJSONA = await fs.readJSON(pkgPathA);
      let pkgJSONB = await fs.readJSON(pkgPathB);

      expect(pkgJSONA).toMatchObject({
        name: "pkg-a",
        version: "1.1.0"
      });
      expect(pkgJSONB).toMatchObject({
        name: "pkg-b",
        version: "2.0.0"
      });
    });
  });
  describe("changelogs", () => {
    it("should update a changelog for one package", async () => {
      const releasePlan = new FakeReleasePlan();
      let { changedFiles } = await testSetup(
        "simple-project",
        releasePlan.getReleasePlan(),
        {
          ...releasePlan.config,
          changelog: [
            path.resolve(__dirname, "test-utils/simple-get-changelog-entry"),
            null
          ]
        }
      );

      let readmePath = changedFiles.find(a => a.endsWith("pkg-a/CHANGELOG.md"));

      if (!readmePath) throw new Error(`could not find an updated changelog`);
      let readme = await fs.readFile(readmePath, "utf-8");

      expect(readme.trim()).toEqual(outdent`# pkg-a

      ## 1.1.0
      ### Minor Changes

      - Hey, let's have fun with testing!`);
    });
    it("should update a changelog for two packages", async () => {
      const releasePlan = new FakeReleasePlan(
        [],
        [
          {
            name: "pkg-b",
            type: "major",
            oldVersion: "1.0.0",
            newVersion: "2.0.0",
            changesets: []
          }
        ]
      );

      let { changedFiles } = await testSetup(
        "simple-project",
        releasePlan.getReleasePlan(),
        {
          ...releasePlan.config,
          changelog: [
            path.resolve(__dirname, "test-utils/simple-get-changelog-entry"),
            null
          ]
        }
      );

      let readmePath = changedFiles.find(a => a.endsWith("pkg-a/CHANGELOG.md"));
      let readmePathB = changedFiles.find(a =>
        a.endsWith("pkg-b/CHANGELOG.md")
      );

      if (!readmePath || !readmePathB)
        throw new Error(`could not find an updated changelog`);
      let readme = await fs.readFile(readmePath, "utf-8");
      let readmeB = await fs.readFile(readmePathB, "utf-8");

      expect(readme.trim()).toEqual(outdent`# pkg-a

      ## 1.1.0
      ### Minor Changes

      - Hey, let's have fun with testing!

      ### Patch Changes

        - pkg-b@2.0.0`);

      expect(readmeB.trim()).toEqual(outdent`# pkg-b

      ## 2.0.0`);
    });

    test("should list multi-line same-type summaries correctly", async () => {
      const releasePlan = new FakeReleasePlan([
        {
          id: "some-id-1",
          summary: "Random stuff\n\nget it while it's hot!",
          releases: [{ name: "pkg-a", type: "minor" }]
        },
        {
          id: "some-id-2",
          summary: "New feature, much wow\n\nlook at this shiny stuff!",
          releases: [{ name: "pkg-a", type: "minor" }]
        }
      ]);
      releasePlan.releases[0].changesets.push("some-id-1", "some-id-2");

      let { changedFiles } = await testSetup(
        "simple-project",
        releasePlan.getReleasePlan(),
        {
          ...releasePlan.config,
          changelog: [
            path.resolve(__dirname, "test-utils/simple-get-changelog-entry"),
            null
          ]
        }
      );

      let readmePath = changedFiles.find(a => a.endsWith("pkg-a/CHANGELOG.md"));

      if (!readmePath) throw new Error(`could not find an updated changelog`);
      let readme = await fs.readFile(readmePath, "utf-8");
      expect(readme.trim()).toEqual(
        [
          "# pkg-a\n",
          "## 1.1.0",
          "### Minor Changes\n",
          "- Hey, let's have fun with testing!",
          "- Random stuff",
          "  \n  get it while it's hot!",
          "- New feature, much wow",
          "  \n  look at this shiny stuff!"
        ].join("\n")
      );
    });
  });
  describe("should error and not write if", () => {
    // This is skipped as *for now* we are assuming we have been passed
    // valid releasePlans - this may get work done on it in the future
    it.skip("a package appears twice", async () => {
      let changedFiles;
      try {
        let testResults = await testSetup("simple-project", {
          changesets: [
            {
              id: "quick-lions-devour",
              summary: "Hey, let's have fun with testing!",
              releases: [{ name: "pkg-a", type: "minor" }]
            }
          ],
          releases: [
            {
              name: "pkg-a",
              type: "minor",
              oldVersion: "1.0.0",
              newVersion: "1.1.0",
              changesets: ["quick-lions-devour"]
            },
            {
              name: "pkg-a",
              type: "minor",
              oldVersion: "1.0.0",
              newVersion: "1.1.0",
              changesets: ["quick-lions-devour"]
            }
          ],
          preState: undefined
        });
        changedFiles = testResults.changedFiles;
      } catch (e) {
        expect(e.message).toEqual("some string probably");

        return;
      }

      throw new Error(
        `expected error but instead got changed files: \n${changedFiles.join(
          "\n"
        )}`
      );
    });
    it("a package cannot be found", async () => {
      let releasePlan = new FakeReleasePlan(
        [],
        [
          {
            name: "impossible-package",
            type: "minor",
            oldVersion: "1.0.0",
            newVersion: "1.0.0",
            changesets: []
          }
        ]
      );

      let tempDir = await copyFixtureIntoTempDir(__dirname, "with-git");

      await spawn("git", ["init"], { cwd: tempDir });

      await git.add(".", tempDir);
      await git.commit("first commit", tempDir);

      try {
        await applyReleasePlan(
          releasePlan.getReleasePlan(),
          await getPackages(tempDir),
          releasePlan.config
        );
      } catch (e) {
        expect(e.message).toEqual(
          "Could not find matching package for release of: impossible-package"
        );

        let gitCmd = await spawn("git", ["status"], { cwd: tempDir });

        expect(gitCmd.stdout.toString().includes("nothing to commit")).toEqual(
          true
        );
        return;
      }

      throw new Error("Expected test to exit before this point");
    });
    it("a provided changelog function fails", async () => {
      let releasePlan = new FakeReleasePlan();

      let tempDir = await copyFixtureIntoTempDir(__dirname, "with-git");

      await spawn("git", ["init"], { cwd: tempDir });

      await git.add(".", tempDir);
      await git.commit("first commit", tempDir);

      try {
        await applyReleasePlan(
          releasePlan.getReleasePlan(),
          await getPackages(tempDir),
          {
            ...releasePlan.config,
            changelog: [
              path.resolve(__dirname, "test-utils/failing-functions"),
              null
            ]
          }
        );
      } catch (e) {
        expect(e.message).toEqual("no chance");

        let gitCmd = await spawn("git", ["status"], { cwd: tempDir });

        expect(gitCmd.stdout.toString().includes("nothing to commit")).toEqual(
          true
        );
        return;
      }

      throw new Error("Expected test to exit before this point");
    });
  });
  describe("changesets", () => {
    it("should delete one changeset after it is applied", async () => {
      const releasePlan = new FakeReleasePlan();

      let changesetPath: string;

      const setupFunc = (tempDir: string) =>
        Promise.all(
          releasePlan.getReleasePlan().changesets.map(({ id, summary }) => {
            const thisPath = path.resolve(tempDir, ".changeset", `${id}.md`);
            changesetPath = thisPath;
            const content = `---\n---\n${summary}`;
            fs.writeFile(thisPath, content);
          })
        );

      await testSetup(
        "simple-project",
        releasePlan.getReleasePlan(),
        releasePlan.config,
        setupFunc
      );

      // @ts-ignore this is possibly bad
      let pathExists = await fs.pathExists(changesetPath);
      expect(pathExists).toEqual(false);
    });
    it("should delete an old format changeset if it is applied", async () => {
      const releasePlan = new FakeReleasePlan();

      let changesetMDPath: string;
      let changesetJSONPath: string;

      const setupFunc = (tempDir: string) =>
        Promise.all(
          releasePlan
            .getReleasePlan()
            .changesets.map(async ({ id, summary }) => {
              changesetMDPath = path.resolve(
                tempDir,
                ".changeset",
                id,
                `changes.md`
              );
              changesetJSONPath = path.resolve(
                tempDir,
                ".changeset",
                id,
                `changes.json`
              );
              await fs.outputFile(changesetMDPath, summary);
              await fs.outputFile(
                changesetJSONPath,
                JSON.stringify({ id, summary })
              );
            })
        );

      await testSetup(
        "simple-project",
        releasePlan.getReleasePlan(),
        releasePlan.config,
        setupFunc
      );

      // @ts-ignore this is possibly bad
      let mdPathExists = await fs.pathExists(changesetMDPath);
      // @ts-ignore this is possibly bad
      let JSONPathExists = await fs.pathExists(changesetMDPath);

      expect(mdPathExists).toEqual(false);
      expect(JSONPathExists).toEqual(false);
    });
  });

  it("should get the commit for an old changeset", async () => {
    const releasePlan = new FakeReleasePlan();

    let changesetMDPath: string;
    let changesetJSONPath: string;

    const setupFunc = (tempDir: string) =>
      Promise.all(
        releasePlan.getReleasePlan().changesets.map(async ({ id, summary }) => {
          changesetMDPath = path.resolve(
            tempDir,
            ".changeset",
            id,
            `changes.md`
          );
          changesetJSONPath = path.resolve(
            tempDir,
            ".changeset",
            id,
            `changes.json`
          );
          await fs.outputFile(changesetMDPath, summary);
          await fs.outputFile(
            changesetJSONPath,
            JSON.stringify({ id, summary })
          );
        })
      );

    let { tempDir } = await testSetup(
      "simple-project",
      releasePlan.getReleasePlan(),
      {
        ...releasePlan.config,
        changelog: [
          path.resolve(__dirname, "test-utils/simple-get-changelog-entry"),
          null
        ],
        commit: true
      },
      setupFunc
    );

    let thing = await spawn("git", ["rev-list", "HEAD"], { cwd: tempDir });
    let commits = thing.stdout
      .toString("utf8")
      .split("\n")
      .filter(x => x);

    let lastCommit = commits[commits.length - 1].substring(0, 7);

    expect(
      await fs.readFile(
        path.join(tempDir, "packages", "pkg-a", "CHANGELOG.md"),
        "utf8"
      )
    ).toBe(`# pkg-a

## 1.1.0
### Minor Changes

- ${lastCommit}: Hey, let's have fun with testing!
`);
  });
  describe("git", () => {
    it("should commit updating files from packages", async () => {
      const releasePlan = new FakeReleasePlan();

      let { tempDir } = await testSetup(
        "with-git",
        releasePlan.getReleasePlan(),
        { ...releasePlan.config, commit: true }
      );

      let gitCmd = await spawn("git", ["status"], { cwd: tempDir });

      expect(gitCmd.stdout.toString().includes("nothing to commit")).toBe(true);

      let lastCommit = await spawn("git", ["log", "-1"], { cwd: tempDir });

      lastCommit.stdout.toString();

      expect(
        lastCommit.stdout
          .toString()
          .includes("RELEASING: Releasing 1 package(s)")
      ).toBe(true);
    });
    it("should commit removing applied changesets", async () => {
      const releasePlan = new FakeReleasePlan();

      let changesetPath: string;

      const setupFunc = (tempDir: string) =>
        Promise.all(
          releasePlan.getReleasePlan().changesets.map(({ id, summary }) => {
            const thisPath = path.resolve(tempDir, ".changeset", `${id}.md`);
            changesetPath = thisPath;
            const content = `---\n---\n${summary}`;
            fs.writeFile(thisPath, content);
          })
        );

      let { tempDir } = await testSetup(
        "with-git",
        releasePlan.getReleasePlan(),
        { ...releasePlan.config, commit: true },
        setupFunc
      );

      // @ts-ignore this is possibly bad
      let pathExists = await fs.pathExists(changesetPath);

      expect(pathExists).toEqual(false);

      let gitCmd = await spawn("git", ["status"], { cwd: tempDir });

      expect(gitCmd.stdout.toString().includes("nothing to commit")).toBe(true);
    });
  });
});

// MAKE SURE BOTH OF THESE ARE COVERED

// it("should git add the expected files (without changelog) when commit: true", async () => {
//   await writeChangesets([simpleChangeset2], cwd);
//   await versionCommand(cwd, { ...modifiedDefaultConfig, commit: true });

//   const pkgAConfigPath = path.join(cwd, "packages/pkg-a/package.json");
//   const pkgBConfigPath = path.join(cwd, "packages/pkg-b/package.json");
//   const changesetConfigPath = path.join(cwd, ".changeset");

//   expect(git.add).toHaveBeenCalledWith(pkgAConfigPath, cwd);
//   expect(git.add).toHaveBeenCalledWith(pkgBConfigPath, cwd);
//   expect(git.add).toHaveBeenCalledWith(changesetConfigPath, cwd);
// });
// it("should git add the expected files (with changelog)", async () => {
//   let changelogPath = path.resolve(__dirname, "../../changelogs");
//   await writeChangesets([simpleChangeset2], cwd);
//   await versionCommand(cwd, {
//     ...modifiedDefaultConfig,
//     changelog: [changelogPath, null],
//     commit: true
//   });
//   const pkgAChangelogPath = path.join(cwd, "packages/pkg-a/CHANGELOG.md");
//   const pkgBChangelogPath = path.join(cwd, "packages/pkg-b/CHANGELOG.md");
//   expect(git.add).toHaveBeenCalledWith(pkgAChangelogPath, cwd);
//   expect(git.add).toHaveBeenCalledWith(pkgBChangelogPath, cwd);
// });
